import json
import re
import time
import urllib.parse as urlparse

import requests

from .confirmation import ConfirmationExecutor
from .exceptions import (
    ApiException,
    InvalidCredentials,
    SevenDaysHoldException,
    SteampieException,
    TooManyRequests,
)
from .guard import load_steam_guard
from .login import LoginExecutor
from .models import Asset, GameOptions, PredefinedOptions, SteamUrl, TradeOfferState
from .utils import (
    account_id_to_steam_id,
    get_description_key,
    get_key_value_from_url,
    login_required,
    merge_items_with_descriptions_from_inventory,
    merge_items_with_descriptions_from_offer,
    merge_items_with_descriptions_from_offers,
    ping_proxy,
    steam_id_to_account_id,
    text_between,
    texts_between,
)


class SteamClient:
    def __init__(
        self,
        api_key: str,
        username: str,
        password: str,
        shared_secret: str,
        identity_secret: str,
        login_cookies: dict | None = None,
        proxies: dict | None = None,
        steam_id: str | None = None,
    ) -> None:
        self._api_key = api_key
        self._session = requests.Session()

        if proxies:
            self.set_proxies(proxies)

        self.was_login_executed = False
        self.username = username
        self._password = password
        self._shared_secret = shared_secret
        self._identity_secret = identity_secret
        self._access_token = None

        # None or str at this point
        self._steam_id = steam_id

        if login_cookies:
            self.set_login_cookies(login_cookies)

    def set_proxies(self, proxies: dict) -> dict:
        if not isinstance(proxies, dict):
            raise TypeError(
                "Proxy must be a dict. Example: "
                + "{'http': 'http://login:password@host:port', "
                + "'https': 'http://login:password@host:port'}",
            )

        if ping_proxy(proxies):
            self._session.proxies.update(proxies)

        return proxies

    def set_login_cookies(self, cookies: dict) -> None:
        self._session.cookies.update(cookies)
        self.was_login_executed = True

    @login_required
    def get_steam_id(self) -> str:
        url = SteamUrl.COMMUNITY_URL
        response = self._session.get(url)

        if steam_id := re.search(r'g_steamID = "(\d+)";', response.text):
            return steam_id.group(1)

        raise ValueError(f"Invalid steam_id: {steam_id}")

    def login(self) -> None:
        # session is alive, no need to login again
        if self.was_login_executed and self.is_session_alive():
            return

        self._session.cookies.set("steamRememberLogin", "true")
        LoginExecutor(
            self.username,
            self._password,
            self._shared_secret,
            self._session,
        ).login()

        self.was_login_executed = True
        self._access_token = self._set_access_token()

        if self._steam_id is None:
            self._steam_id = self.get_steam_id()

    def _set_access_token(self) -> str:
        steam_login_secure_cookies = [
            cookie
            for cookie in self._session.cookies
            if cookie.name == "steamLoginSecure"
        ]
        cookie_value = steam_login_secure_cookies[0].value
        decoded_cookie_value = urlparse.unquote(cookie_value)
        access_token_parts = decoded_cookie_value.split("||")

        if len(access_token_parts) < 2:
            print(decoded_cookie_value)
            raise ValueError("Access token not found in steamLoginSecure cookie")

        access_token = access_token_parts[1]
        return access_token

    @login_required
    def logout(self) -> None:
        url = f"{SteamUrl.COMMUNITY_URL}/login/logout/"
        data = {"sessionid": self._get_session_id()}
        self._session.post(url, data=data)

        if self.is_session_alive():
            raise SteampieException("Logout unsuccessful")

        self.was_login_executed = False

    def __enter__(self):
        self.login()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del exc_type, exc_val, exc_tb
        self.logout()

    @login_required
    def is_session_alive(self) -> bool:
        steam_login = self.username
        main_page_response = self._session.get(SteamUrl.COMMUNITY_URL)
        return steam_login.lower() in main_page_response.text.lower()

    def api_call(
        self,
        method: str,
        interface: str,
        api_method: str,
        version: str,
        params: dict | None = None,
    ) -> requests.Response:
        url = f"{SteamUrl.API_URL}/{interface}/{api_method}/{version}"
        response = None

        if method == "GET":
            response = self._session.get(url, params=params)
        else:
            response = self._session.post(url, data=params)

        if self.is_invalid_api_key(response):
            raise InvalidCredentials("Invalid API key")

        return response

    @staticmethod
    def is_invalid_api_key(response: requests.Response) -> bool:
        msg = "Access is denied. Retrying will not help. Please verify your <pre>key=</pre> parameter"
        return msg in response.text

    @login_required
    def get_my_inventory(
        self, game: GameOptions | str, merge: bool = True, count: int = 5000
    ) -> dict:
        return self.get_partner_inventory(self.steam_id, game, merge, count)

    @login_required
    def get_partner_inventory(
        self,
        partner_steam_id: str,
        game: GameOptions | str,
        context_id: str = "2",
        merge: bool = True,
        count: int = 5000,
    ) -> dict:
        app_id = game
        context_id = context_id

        if isinstance(game, PredefinedOptions):
            app_id = game.app_id
            context_id = game.context_id

        if not isinstance(app_id, str):
            raise SteampieException("Invalid app_id type")

        url = f"{SteamUrl.COMMUNITY_URL}/inventory/{partner_steam_id}/{app_id}/{context_id}"
        params = {"l": "english", "count": count}

        full_response = self._session.get(url, params=params)
        response_dict = full_response.json()

        if full_response.status_code == 429:
            raise TooManyRequests("Too many requests, try again later.")

        if response_dict is None or response_dict.get("success") != 1:
            raise ApiException("Success value should be 1.")

        return (
            merge_items_with_descriptions_from_inventory(response_dict, game)
            if merge
            else response_dict
        )

    def _get_session_id(self) -> str | None:
        response = self._session.cookies.get_dict(domain="steamcommunity.com", path="/")
        return response.get("sessionid")

    def get_trade_offers_summary(self) -> dict:
        params = {"key": self._api_key}
        response = self.api_call(
            "GET", "IEconService", "GetTradeOffersSummary", "v1", params
        )
        return response.json()

    def _set_token_params(self, params: dict, use_webtoken: bool) -> dict:
        if use_webtoken:
            params["access_token"] = self._access_token
        else:
            params["key"] = self._api_key

        return params

    def get_trade_offers(
        self,
        merge: bool = True,
        get_sent_offers: bool = True,
        get_received_offers: bool = True,
        use_webtoken: bool = False,
        max_retry: int = 5,
    ) -> dict:
        params = {
            "get_sent_offers": int(get_sent_offers),
            "get_received_offers": int(get_received_offers),
            "get_descriptions": 1,
            "language": "english",
            "active_only": 1,
            "historical_only": 0,
            "time_historical_cutoff": "",
        }

        params = self._set_token_params(params, use_webtoken)
        response = self._try_to_get_trade_offers(params, max_retry)

        if response is None:
            raise ApiException("Cannot get proper json from get_trade_offers method")

        response_with_active_offers = self._filter_non_active_offers(response)

        if not merge:
            return response_with_active_offers

        return merge_items_with_descriptions_from_offers(response_with_active_offers)

    def _try_to_get_trade_offers(self, params: dict, max_retry: int) -> dict | None:
        response = None

        for _ in range(max_retry):
            try:
                response = self.api_call(
                    "GET", "IEconService", "GetTradeOffers", "v1", params
                ).json()
                break
            except json.decoder.JSONDecodeError:
                time.sleep(2)

        return response

    @staticmethod
    def _filter_non_active_offers(offers_response):
        offers_received = offers_response["response"].get("trade_offers_received", [])
        offers_sent = offers_response["response"].get("trade_offers_sent", [])

        offers_response["response"]["trade_offers_received"] = list(
            filter(
                lambda offer: offer["trade_offer_state"] == TradeOfferState.Active,
                offers_received,
            ),
        )
        offers_response["response"]["trade_offers_sent"] = list(
            filter(
                lambda offer: offer["trade_offer_state"] == TradeOfferState.Active,
                offers_sent,
            ),
        )

        return offers_response

    def get_trade_offer(
        self, trade_offer_id: str, merge: bool = True, use_webtoken: bool = False
    ) -> dict:
        params = {"tradeofferid": trade_offer_id, "language": "english"}
        params = self._set_token_params(params, use_webtoken)

        response = self.api_call(
            "GET", "IEconService", "GetTradeOffer", "v1", params
        ).json()

        if merge and "descriptions" in response["response"]:
            descriptions = {
                get_description_key(offer): offer
                for offer in response["response"]["descriptions"]
            }
            offer = response["response"]["offer"]
            response["response"]["offer"] = merge_items_with_descriptions_from_offer(
                offer, descriptions
            )

        return response

    def get_trade_history(
        self,
        max_trades: int = 100,
        start_after_time=None,
        start_after_tradeid=None,
        get_descriptions: bool = True,
        navigating_back: bool = True,
        include_failed: bool = True,
        include_total: bool = True,
    ) -> dict:
        params = {
            "key": self._api_key,
            "max_trades": max_trades,
            "start_after_time": start_after_time,
            "start_after_tradeid": start_after_tradeid,
            "get_descriptions": get_descriptions,
            "navigating_back": navigating_back,
            "include_failed": include_failed,
            "include_total": include_total,
        }
        response = self.api_call("GET", "IEconService", "GetTradeHistory", "v1", params)
        return response.json()

    @login_required
    def get_trade_receipt(self, trade_id: str):
        html = self._session.get(
            f"https://steamcommunity.com/trade/{trade_id}/receipt"
        ).content.decode()
        return [
            json.loads(item) for item in texts_between(html, "oItem = ", ";\r\n\toItem")
        ]

    @login_required
    def accept_trade_offer(self, trade_offer_id: str) -> dict:
        trade = self.get_trade_offer(trade_offer_id, use_webtoken=True)
        trade_offer_state = TradeOfferState(
            trade["response"]["offer"]["trade_offer_state"]
        )

        if trade_offer_state is not TradeOfferState.Active:
            raise ApiException(
                f"Invalid trade offer state: {trade_offer_state.name} ({trade_offer_state.value})"
            )

        partner = self._fetch_trade_partner_id(trade_offer_id)
        session_id = self._get_session_id()
        accept_url = f"{SteamUrl.COMMUNITY_URL}/tradeoffer/{trade_offer_id}/accept"
        params = {
            "sessionid": session_id,
            "tradeofferid": trade_offer_id,
            "serverid": "1",
            "partner": partner,
            "captcha": "",
        }
        headers = {"Referer": self._get_trade_offer_url(trade_offer_id)}

        response = self._session.post(accept_url, data=params, headers=headers).json()

        if response.get("needs_mobile_confirmation", False):
            return self._confirm_transaction(trade_offer_id)

        return response

    def _fetch_trade_partner_id(self, trade_offer_id: str) -> str:
        url = self._get_trade_offer_url(trade_offer_id)
        offer_response_text = self._session.get(url).text

        if (
            "You have logged in from a new device. In order to protect the items"
            in offer_response_text
        ):
            raise SevenDaysHoldException(
                "Account has logged in a new device and can't trade for 7 days"
            )

        return text_between(
            offer_response_text, "var g_ulTradePartnerSteamID = '", "';"
        )

    def _confirm_transaction(self, trade_offer_id: str) -> dict:
        confirmation_executor = ConfirmationExecutor(
            self._identity_secret,
            self.steam_id,
            self._session,
        )
        return confirmation_executor.send_trade_allow_request(trade_offer_id)

    def decline_trade_offer(self, trade_offer_id: str) -> dict:
        url = f"https://steamcommunity.com/tradeoffer/{trade_offer_id}/decline"
        response = self._session.post(url, data={"sessionid": self._get_session_id()})
        return response.json()

    def cancel_trade_offer(self, trade_offer_id: str) -> dict:
        url = f"https://steamcommunity.com/tradeoffer/{trade_offer_id}/cancel"
        response = self._session.post(url, data={"sessionid": self._get_session_id()})
        return response.json()

    @login_required
    def make_offer(
        self,
        items_from_me: list[Asset],
        items_from_them: list[Asset],
        partner_steam_id: str,
        message: str = "",
    ) -> dict:
        offer = self._create_offer_dict(items_from_me, items_from_them)
        session_id = self._get_session_id()
        url = f"{SteamUrl.COMMUNITY_URL}/tradeoffer/new/send"
        server_id = 1
        params = {
            "sessionid": session_id,
            "serverid": server_id,
            "partner": partner_steam_id,
            "tradeoffermessage": message,
            "json_tradeoffer": json.dumps(offer),
            "captcha": "",
            "trade_offer_create_params": "{}",
        }
        partner_account_id = steam_id_to_account_id(partner_steam_id)
        headers = {
            "Referer": f"{SteamUrl.COMMUNITY_URL}/tradeoffer/new/?partner={partner_account_id}",
            "Origin": SteamUrl.COMMUNITY_URL,
        }
        response = self._session.post(url, data=params, headers=headers).json()

        if response.get("needs_mobile_confirmation"):
            response.update(self._confirm_transaction(response["tradeofferid"]))

        return response

    def get_profile(self, steam_id: str) -> dict:
        params = {"steamids": steam_id, "key": self._api_key}
        response = self.api_call(
            "GET", "ISteamUser", "GetPlayerSummaries", "v0002", params
        )
        data = response.json()
        return data["response"]["players"][0]

    def get_friend_list(self, steam_id: str, relationship_filter: str = "all") -> dict:
        params = {
            "key": self._api_key,
            "steamid": steam_id,
            "relationship": relationship_filter,
        }
        resp = self.api_call("GET", "ISteamUser", "GetFriendList", "v1", params)
        data = resp.json()
        return data["friendslist"]["friends"]

    @staticmethod
    def _create_offer_dict(
        items_from_me: list[Asset], items_from_them: list[Asset]
    ) -> dict:
        return {
            "newversion": True,
            "version": 4,
            "me": {
                "assets": [asset.to_dict() for asset in items_from_me],
                "currency": [],
                "ready": False,
            },
            "them": {
                "assets": [asset.to_dict() for asset in items_from_them],
                "currency": [],
                "ready": False,
            },
        }

    @login_required
    def get_escrow_duration(self, trade_offer_url: str) -> int:
        headers = {
            "Referer": f"{SteamUrl.COMMUNITY_URL}{urlparse.urlparse(trade_offer_url).path}",
            "Origin": SteamUrl.COMMUNITY_URL,
        }
        response = self._session.get(trade_offer_url, headers=headers).text

        my_escrow_duration = int(text_between(response, "var g_daysMyEscrow = ", ";"))
        their_escrow_duration = int(
            text_between(response, "var g_daysTheirEscrow = ", ";")
        )

        return max(my_escrow_duration, their_escrow_duration)

    @login_required
    def make_offer_with_url(
        self,
        items_from_me: list[Asset],
        items_from_them: list[Asset],
        trade_offer_url: str,
        message: str = "",
        case_sensitive: bool = True,
        confirm_trade: bool = True,
    ) -> dict:
        token = get_key_value_from_url(trade_offer_url, "token", case_sensitive)
        partner_account_id = get_key_value_from_url(
            trade_offer_url, "partner", case_sensitive
        )
        partner_steam_id = account_id_to_steam_id(partner_account_id)
        offer = self._create_offer_dict(items_from_me, items_from_them)
        session_id = self._get_session_id()
        url = f"{SteamUrl.COMMUNITY_URL}/tradeoffer/new/send"
        server_id = 1
        trade_offer_create_params = {"trade_offer_access_token": token}
        params = {
            "sessionid": session_id,
            "serverid": server_id,
            "partner": partner_steam_id,
            "tradeoffermessage": message,
            "json_tradeoffer": json.dumps(offer),
            "captcha": "",
            "trade_offer_create_params": json.dumps(trade_offer_create_params),
        }

        headers = {
            "Referer": f"{SteamUrl.COMMUNITY_URL}{urlparse.urlparse(trade_offer_url).path}",
            "Origin": SteamUrl.COMMUNITY_URL,
        }

        response = self._session.post(url, data=params, headers=headers).json()

        if confirm_trade and response.get("needs_mobile_confirmation"):
            response.update(self._confirm_transaction(response["tradeofferid"]))

        return response

    def _friend_ajax_request(self, method: str, steam_id: str, accept: int = 0) -> dict:
        assert method.lower() in ["add", "remove"]

        headers = {"Content-Type": "application/x-www-form-urlencoded; charset=UTF-8"}
        url = f"{SteamUrl.COMMUNITY_URL}/actions/AddFriendAjax"
        data = {
            "sessionID": self._get_session_id(),
            "steamid": steam_id,
            "accept_invite": accept,
        }

        if method == "remove":
            url = f"{SteamUrl.COMMUNITY_URL}/actions/RemoveFriendAjax"
            del data["accept_invite"]

        response = self._session.post(url, data=data, headers=headers)

        if response.status_code != 200:
            raise ApiException(
                f"Request failed with status code {response.status_code}"
            )

        return response.json()

    @login_required
    def remove_friend(self, steam_id: str) -> dict:
        return self._friend_ajax_request("remove", steam_id)

    @login_required
    def add_friend(self, steam_id: str) -> dict:
        return self._friend_ajax_request("add", steam_id)

    @login_required
    def accept_friend(self, steam_id: str) -> dict:
        return self._friend_ajax_request("add", steam_id, accept=1)

    @staticmethod
    def _get_trade_offer_url(trade_offer_id: str) -> str:
        return f"{SteamUrl.COMMUNITY_URL}/tradeoffer/{trade_offer_id}"

    @property
    def steam_id(self) -> str:
        return self._steam_id


class SteamGuardLoginClient(SteamClient):
    def __init__(
        self,
        api_key: str,
        password: str,
        steam_guard_file: str,
    ) -> None:
        steam_guard = load_steam_guard(steam_guard_file)
        username = steam_guard["account_name"]
        shared_secret = steam_guard["shared_secret"]
        identity_secret = steam_guard["identity_secret"]
        steam_id = str(steam_guard["Session"]["SteamID"])

        super().__init__(
            api_key=api_key,
            username=username,
            password=password,
            shared_secret=shared_secret,
            identity_secret=identity_secret,
            steam_id=steam_id,
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
# flake8: noqa
import re
import os
import sys

sys.path.insert(0, os.path.abspath(".."))
sys.path.append(os.path.abspath("extensions"))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "steampie"
copyright = "2024, offish"
author = "offish"

with open("src/steampie/__init__.py") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(1)

# The full version, including alpha/beta/rc tags
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

extensions = [
    "sphinxcontrib.jquery",
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_rtd_theme",
    "sphinx.ext.napoleon",
]

autodoc_member_order = "bysource"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

html_theme = "furo"



from steampie.confirmation import Tag
from steampie.guard import (
    generate_confirmation_key,
    generate_device_id,
    generate_one_time_code,
    load_steam_guard,
)


def test_one_time_code(shared_secret) -> None:
    timestamp = 1469184207
    code = generate_one_time_code(shared_secret, timestamp)
    assert code == "P2QJN"


def test_confirmation_key(identity_secret) -> None:
    timestamp = 1470838334
    confirmation_key = generate_confirmation_key(
        identity_secret, Tag.CONF.value, timestamp
    )
    assert confirmation_key == b"pWqjnkcwqni+t/n+5xXaEa0SGeA="


def test_generate_device_id() -> None:
    steam_id = "12341234123412345"
    device_id = generate_device_id(steam_id)
    assert device_id == "android:677cf5aa-3300-7807-d1e2-c408142742e2"


def test_load_steam_guard() -> None:
    guard_json_str = '{"steamid": 12345678, "shared_secret": "SHARED_SECRET", "identity_secret": "IDENTITY_SECRET"}'
    guard_data = load_steam_guard(guard_json_str)

    for key in ["steamid", "shared_secret", "identity_secret"]:
        assert key in guard_data
        assert isinstance(guard_data[key], str)



import struct
from copy import copy
from urllib.parse import parse_qs, urlparse

import requests
from requests.structures import CaseInsensitiveDict

from .exceptions import LoginRequired, ProxyConnectionError
from .models import GameOptions

__all__ = [
    "login_required",
    "text_between",
    "texts_between",
    "account_id_to_steam_id",
    "steam_id_to_account_id",
    "merge_items_with_descriptions_from_inventory",
    "merge_items_with_descriptions_from_offers",
    "merge_items_with_descriptions_from_offer",
    "merge_items",
    "get_description_key",
    "get_key_value_from_url",
    "ping_proxy",
    "create_cookie",
]


class Credentials:
    def __init__(self, login: str, password: str, api_key: str) -> None:
        self.login = login
        self.password = password
        self.api_key = api_key


def login_required(func):
    def func_wrapper(self, *args, **kwargs):
        if not self.was_login_executed:
            raise LoginRequired("Use login method first")
        return func(self, *args, **kwargs)

    return func_wrapper


def text_between(text: str, begin: str, end: str) -> str:
    start = text.index(begin) + len(begin)
    end = text.index(end, start)

    return text[start:end]


def texts_between(text: str, begin: str, end: str):
    stop = 0

    while True:
        try:
            start = text.index(begin, stop) + len(begin)
            stop = text.index(end, start)
            yield text[start:stop]
        except ValueError:
            return


def account_id_to_steam_id(account_id: str) -> str:
    first_bytes = int(account_id).to_bytes(4, byteorder="big")
    last_bytes = 0x1100001.to_bytes(4, byteorder="big")
    return str(struct.unpack(">Q", last_bytes + first_bytes)[0])


def steam_id_to_account_id(steam_id: str) -> str:
    return str(struct.unpack(">L", int(steam_id).to_bytes(8, byteorder="big")[4:])[0])


def merge_items_with_descriptions_from_inventory(
    inventory_response: dict, game: GameOptions
) -> dict:
    inventory = inventory_response.get("assets", [])

    if not inventory:
        return {}

    descriptions = {
        get_description_key(description): description
        for description in inventory_response["descriptions"]
    }
    return merge_items(inventory, descriptions, context_id=game.context_id)


def merge_items_with_descriptions_from_offers(offers_response: dict) -> dict:
    descriptions = {
        get_description_key(offer): offer
        for offer in offers_response["response"].get("descriptions", [])
    }

    received_offers = offers_response["response"].get("trade_offers_received", [])
    sent_offers = offers_response["response"].get("trade_offers_sent", [])

    offers_response["response"]["trade_offers_received"] = [
        merge_items_with_descriptions_from_offer(offer, descriptions)
        for offer in received_offers
    ]

    offers_response["response"]["trade_offers_sent"] = [
        merge_items_with_descriptions_from_offer(offer, descriptions)
        for offer in sent_offers
    ]

    return offers_response


def merge_items_with_descriptions_from_offer(offer: dict, descriptions: dict) -> dict:
    merged_items_to_give = merge_items(offer.get("items_to_give", []), descriptions)
    merged_items_to_receive = merge_items(
        offer.get("items_to_receive", []), descriptions
    )

    offer["items_to_give"] = merged_items_to_give
    offer["items_to_receive"] = merged_items_to_receive

    return offer


def merge_items(items: list[dict], descriptions: dict, **kwargs) -> dict:
    merged_items = {}

    for item in items:
        description_key = get_description_key(item)
        description = copy(descriptions[description_key])
        item_id = item.get("id") or item["assetid"]
        description["contextid"] = item.get("contextid") or kwargs["context_id"]
        description["id"] = item_id
        description["amount"] = item["amount"]
        merged_items[item_id] = description

    return merged_items


def get_description_key(item: dict) -> str:
    return f'{item["classid"]}_{item["instanceid"]}'


def get_key_value_from_url(url: str, key: str, case_sensitive: bool = True) -> str:
    params = urlparse(url).query
    return (
        parse_qs(params)[key][0]
        if case_sensitive
        else CaseInsensitiveDict(parse_qs(params))[key][0]
    )


def ping_proxy(proxies: dict) -> bool:
    try:
        requests.get("https://steamcommunity.com/", proxies=proxies)
        return True
    except Exception:
        raise ProxyConnectionError("Proxy not working for steamcommunity.com")


def create_cookie(name: str, cookie: str, domain: str) -> dict:
    return {"name": name, "value": cookie, "domain": domain}



from steampie.utils import (
    account_id_to_steam_id,
    get_key_value_from_url,
    steam_id_to_account_id,
    text_between,
    texts_between,
)


def test_text_between() -> None:
    text = 'var a = "dupadupa";'
    assert text_between(text, 'var a = "', '";') == "dupadupa"


def test_texts_between() -> None:
    text = "<li>element 1</li>\n<li>some random element</li>"
    items = list(texts_between(text, "<li>", "</li>"))
    assert items == ["element 1", "some random element"]


def test_account_id_to_steam_id() -> None:
    assert account_id_to_steam_id("293059984") == "76561198253325712"
    assert account_id_to_steam_id("488991480") == "76561198449257208"


def test_steam_id_to_account_id() -> None:
    assert steam_id_to_account_id("76561198253325712") == "293059984"
    assert steam_id_to_account_id("76561198449257208") == "488991480"


def test_get_key_value_from_url() -> None:
    url = "https://steamcommunity.com/tradeoffer/new/?partner=aaa&token=bbb"
    assert get_key_value_from_url(url, "partner") == "aaa"
    assert get_key_value_from_url(url, "token") == "bbb"


def test_get_key_value_from_url_case_insensitive() -> None:
    url = "https://steamcommunity.com/tradeoffer/new/?Partner=aaa&Token=bbb"
    assert get_key_value_from_url(url, "partner", case_sensitive=False) == "aaa"
    assert get_key_value_from_url(url, "token", case_sensitive=False) == "bbb"



__all__ = [
    "SteampieException",
    "SevenDaysHoldException",
    "TooManyRequests",
    "ApiException",
    "LoginRequired",
    "InvalidCredentials",
    "CaptchaRequired",
    "ConfirmationExpected",
    "ProxyConnectionError",
]


class SteampieException(Exception):
    pass


class SevenDaysHoldException(SteampieException):
    pass


class TooManyRequests(SteampieException):
    pass


class ApiException(SteampieException):
    pass


class LoginRequired(SteampieException):
    pass


class InvalidCredentials(SteampieException):
    pass


class CaptchaRequired(SteampieException):
    pass


class ConfirmationExpected(SteampieException):
    pass


class ProxyConnectionError(SteampieException):
    pass



import json
from base64 import b64encode

import pytest

from steampie.client import SteamClient


def my_steam_id() -> str:
    return "76561198253325712"


@pytest.fixture
def partner_steam_id() -> str:
    return "76561198155938401"


@pytest.fixture
def partner_trade_url() -> str:
    return "https://steamcommunity.com/tradeoffer/new/?partner=195672673&token=k3YfpXNU"


@pytest.fixture
def shared_secret() -> bytes:
    return b64encode(b"1234567890abcdefghij")


@pytest.fixture
def identity_secret() -> bytes:
    return b64encode(b"abcdefghijklmnoprstu")


def get_credentials() -> dict:
    steam_credentials = {}

    with open("./credentials.json", "r") as f:
        steam_credentials = json.load(f)

    return steam_credentials


@pytest.fixture
def credentials() -> dict:
    return get_credentials()


@pytest.fixture
def steam_guard_file() -> dict:
    steam_guard_credentials = {}
    steam_id = my_steam_id()

    with open(f"./{steam_id}.maFile", "r") as f:
        steam_guard_credentials = json.load(f)

    steam_credentials = get_credentials()

    steam_guard_credentials["password"] = steam_credentials["password"]
    steam_guard_credentials["api_key"] = steam_credentials["api_key"]

    return steam_guard_credentials


steam_credentials = get_credentials()
steam_client = SteamClient(**steam_credentials)
steam_client.login()


@pytest.fixture
def client() -> SteamClient:
    return steam_client



from enum import IntEnum
from typing import NamedTuple


class PredefinedOptions(NamedTuple):
    app_id: str
    context_id: str


class GameOptions:
    STEAM = PredefinedOptions("753", "6")
    DOTA2 = PredefinedOptions("570", "2")
    CS = PredefinedOptions("730", "2")
    TF2 = PredefinedOptions("440", "2")
    RUST = PredefinedOptions("252490", "2")

    def __init__(self, app_id: str, context_id: str) -> None:
        self.app_id = app_id
        self.context_id = context_id


class Asset:
    def __init__(self, asset_id: str, game: GameOptions, amount: int = 1) -> None:
        self.asset_id = asset_id
        self.game = game
        self.amount = amount

    def to_dict(self) -> dict:
        return {
            "appid": int(self.game.app_id),
            "contextid": self.game.context_id,
            "amount": self.amount,
            "assetid": self.asset_id,
        }


class Currency(IntEnum):
    USD = 1
    GBP = 2
    EUR = 3
    CHF = 4
    RUB = 5
    PLN = 6
    BRL = 7
    JPY = 8
    NOK = 9
    IDR = 10
    MYR = 11
    PHP = 12
    SGD = 13
    THB = 14
    VND = 15
    KRW = 16
    # TRY = 17
    UAH = 18
    MXN = 19
    CAD = 20
    AUD = 21
    NZD = 22
    CNY = 23
    INR = 24
    CLP = 25
    PEN = 26
    COP = 27
    ZAR = 28
    HKD = 29
    TWD = 30
    SAR = 31
    AED = 32
    # SEK = 33
    # ARS = 34
    ILS = 35
    # BYN = 36
    KZT = 37
    KWD = 38
    QAR = 39
    CRC = 40
    UYU = 41
    # BGN = 42
    # HRK = 43
    # CZK = 44
    # DKK = 45
    # HUF = 46
    # RON = 47


class TradeOfferState(IntEnum):
    Invalid = 1
    Active = 2
    Accepted = 3
    Countered = 4
    Expired = 5
    Canceled = 6
    Declined = 7
    InvalidItems = 8
    ConfirmationNeed = 9
    CanceledBySecondaryFactor = 10
    StateInEscrow = 11


class SteamUrl:
    API_URL = "https://api.steampowered.com"
    COMMUNITY_URL = "https://steamcommunity.com"
    STORE_URL = "https://store.steampowered.com"
    LOGIN_URL = "https://login.steampowered.com"


class Endpoints:
    CHAT_LOGIN = f"{SteamUrl.API_URL}/ISteamWebUserPresenceOAuth/Logon/v1"
    SEND_MESSAGE = f"{SteamUrl.API_URL}/ISteamWebUserPresenceOAuth/Message/v1"
    CHAT_LOGOUT = f"{SteamUrl.API_URL}/ISteamWebUserPresenceOAuth/Logoff/v1"
    CHAT_POLL = f"{SteamUrl.API_URL}/ISteamWebUserPresenceOAuth/Poll/v1"



import pytest

from steampie.client import SteamClient
from steampie.exceptions import LoginRequired
from steampie.models import Asset, GameOptions


def test_steam_login(client: SteamClient) -> None:
    assert client.is_session_alive()


def test_get_steam_id(client: SteamClient, steam_guard_file) -> None:
    assert client.steam_id == str(steam_guard_file["Session"]["SteamID"])


def test_sessionid_cookie(client: SteamClient) -> None:
    community_cookies = client._session.cookies.get_dict("steamcommunity.com")
    store_cookies = client._session.cookies.get_dict("store.steampowered.com")

    assert "sessionid" in community_cookies
    assert "sessionid" in store_cookies


def test_get_my_inventory(client: SteamClient) -> None:
    inventory = client.get_my_inventory(GameOptions.CS)
    assert inventory is not None


def test_get_partner_inventory(client: SteamClient) -> None:
    partner_id = "76561198449257208"
    game = GameOptions.TF2
    inventory = client.get_partner_inventory(partner_id, game)
    assert inventory is not None


def test_get_trade_offers_summary(client: SteamClient) -> None:
    summary = client.get_trade_offers_summary()
    assert summary is not None


def test_get_trade_offers(client: SteamClient) -> None:
    offers = client.get_trade_offers()
    assert offers is not None


# def test_get_trade_offer() -> None:
#     trade_offer_id = "488991480"
#     offer = client.get_trade_offer(trade_offer_id)
#     assert offer is not None


# def test_accept_trade_offer() -> None:
#     trade_offer_id = "1451378159"
#     response_dict = client.accept_trade_offer(trade_offer_id)
#     assert response_dict is not None


# def test_decline_trade_offer() -> None:
#     trade_offer_id = "1449530707"
#     response_dict = client.decline_trade_offer(trade_offer_id)
#     assert response_dict["response"] == {}


# def test_cancel_trade_offer() -> None:
#     trade_offer_id = "1450637835"
#     response_dict = client.cancel_trade_offer(trade_offer_id)
#     assert response_dict["response"] == {}


def test_make_offer_url(
    client: SteamClient, partner_steam_id: str, partner_trade_url: str
) -> None:
    game = GameOptions.TF2
    my_items = client.get_my_inventory(GameOptions.TF2)
    partner_items = client.get_partner_inventory(partner_steam_id, game)
    my_first_item = next(iter(my_items.values()))
    partner_first_item = next(iter(partner_items.values()))
    my_asset = Asset(my_first_item["id"], game)
    partner_asset = Asset(partner_first_item["id"], game)
    response = client.make_offer_with_url(
        [my_asset], [partner_asset], partner_trade_url, "test offer"
    )

    assert response is not None
    assert "tradeofferid" in response


def test_make_offer(client: SteamClient, partner_steam_id: str) -> None:
    game = GameOptions.TF2
    my_items = client.get_my_inventory(GameOptions.TF2)
    partner_items = client.get_partner_inventory(partner_steam_id, game)
    my_first_item = next(iter(my_items.values()))
    partner_first_item = next(iter(partner_items.values()))
    my_asset = Asset(my_first_item["id"], game)
    partner_asset = Asset(partner_first_item["id"], game)
    response = client.make_offer(
        [my_asset], [partner_asset], partner_steam_id, "test offer"
    )

    assert response is not None
    assert "tradeofferid" in response


def test_get_escrow_duration(client: SteamClient, partner_trade_url: str) -> None:
    response = client.get_escrow_duration(partner_trade_url)
    assert response == 0


def test_logout(client: SteamClient) -> None:
    assert client.is_session_alive()
    client.logout()
    assert not client.was_login_executed


def test_accept_trade_offer_without_login(credentials: dict) -> None:
    client = SteamClient(credentials["api_key"], None, None, None, None)

    with pytest.raises(LoginRequired):
        client.accept_trade_offer("id")



# flake8: noqa
__title__ = "steampie"
__author__ = "offish"
__version__ = "0.2.0"
__license__ = "MIT"

from .client import SteamClient, SteamGuardLoginClient
from .exceptions import *
from .models import Asset, Currency, GameOptions, TradeOfferState
from .utils import *



import hmac
import json
import struct
from base64 import b64decode, b64encode
from hashlib import sha1
from pathlib import Path
from time import time


def load_steam_guard(steam_guard: str) -> dict[str, str]:
    """Load Steam Guard credentials from json (file or string).

    Arguments:
        steam_guard: If this string is a path to a file, then its contents will be parsed as a json data.
            Otherwise, the string will be parsed as a json data.

    Returns:
        Parsed json data as a dictionary of strings (both key and value).

    """
    if Path(steam_guard).is_file():
        with Path(steam_guard).open(encoding="utf-8") as f:
            return json.loads(f.read(), parse_int=str)
    else:
        return json.loads(steam_guard, parse_int=str)


def generate_one_time_code(shared_secret: str, timestamp: int | None = None) -> str:
    if timestamp is None:
        timestamp = int(time())

    # pack as Big endian, uint64
    time_buffer = struct.pack(">Q", timestamp // 30)
    time_hmac = hmac.new(b64decode(shared_secret), time_buffer, digestmod=sha1).digest()
    begin = ord(time_hmac[19:20]) & 0xF
    # unpack as Big endian uint32
    full_code = struct.unpack(">I", time_hmac[begin : begin + 4])[0] & 0x7FFFFFFF
    chars = "23456789BCDFGHJKMNPQRTVWXY"
    code = ""

    for _ in range(5):
        full_code, i = divmod(full_code, len(chars))
        code += chars[i]

    return code


def generate_confirmation_key(
    identity_secret: str, tag: str, timestamp: int = int(time())
) -> bytes:
    buffer = struct.pack(">Q", timestamp) + tag.encode("ascii")
    return b64encode(
        hmac.new(b64decode(identity_secret), buffer, digestmod=sha1).digest()
    )


# It works, however it's different that one generated from mobile app
def generate_device_id(steam_id: str) -> str:
    hexed_steam_id = sha1(steam_id.encode("ascii")).hexdigest()
    return "android:" + "-".join(
        (
            hexed_steam_id[:8],
            hexed_steam_id[8:12],
            hexed_steam_id[12:16],
            hexed_steam_id[16:20],
            hexed_steam_id[20:32],
        )
    )



import enum
import json
import time
from http import HTTPStatus

from bs4 import BeautifulSoup
from requests import Response, Session

from .exceptions import ConfirmationExpected, InvalidCredentials
from .guard import generate_confirmation_key, generate_device_id


class Confirmation:
    def __init__(self, data_confid, nonce) -> None:
        self.data_confid = data_confid
        self.nonce = nonce


class Tag(enum.Enum):
    CONF = "conf"
    DETAILS = "details"
    ALLOW = "allow"
    CANCEL = "cancel"


class ConfirmationExecutor:
    CONF_URL = "https://steamcommunity.com/mobileconf"

    def __init__(
        self, identity_secret: str, my_steam_id: str, session: Session
    ) -> None:
        self._my_steam_id = my_steam_id
        self._identity_secret = identity_secret
        self._session = session

    def send_trade_allow_request(self, trade_offer_id: str) -> dict:
        confirmations = self._get_confirmations()
        confirmation = self._select_trade_offer_confirmation(
            confirmations, trade_offer_id
        )

        return self._send_confirmation(confirmation)

    def _send_confirmation(self, confirmation: Confirmation) -> dict:
        tag = Tag.ALLOW
        params = self._create_confirmation_params(tag.value)
        params["op"] = (tag.value,)
        params["cid"] = confirmation.data_confid
        params["ck"] = confirmation.nonce
        headers = {"X-Requested-With": "XMLHttpRequest"}

        return self._session.get(
            f"{self.CONF_URL}/ajaxop", params=params, headers=headers
        ).json()

    def _get_confirmations(self) -> list[Confirmation]:
        confirmations = []
        confirmations_page = self._fetch_confirmations_page()

        if confirmations_page.status_code == HTTPStatus.OK:
            confirmations_json = json.loads(confirmations_page.text)
            for conf in confirmations_json["conf"]:
                data_confid = conf["id"]
                nonce = conf["nonce"]
                confirmations.append(Confirmation(data_confid, nonce))
            return confirmations

        raise ConfirmationExpected

    def _fetch_confirmations_page(self) -> Response:
        tag = Tag.CONF.value
        params = self._create_confirmation_params(tag)
        headers = {"X-Requested-With": "com.valvesoftware.android.steam.community"}
        response = self._session.get(
            f"{self.CONF_URL}/getlist", params=params, headers=headers
        )
        if (
            "Steam Guard Mobile Authenticator is providing incorrect Steam Guard codes."
            in response.text
        ):
            raise InvalidCredentials("Invalid Steam Guard file")

        return response

    def _fetch_confirmation_details_page(self, confirmation: Confirmation) -> str:
        tag = f"details{confirmation.data_confid}"
        params = self._create_confirmation_params(tag)
        response = self._session.get(
            f"{self.CONF_URL}/details/{confirmation.data_confid}", params=params
        )

        return response.json()["html"]

    def _create_confirmation_params(self, tag_string: str) -> dict:
        timestamp = int(time.time())
        confirmation_key = generate_confirmation_key(
            self._identity_secret, tag_string, timestamp
        )
        android_id = generate_device_id(self._my_steam_id)

        return {
            "p": android_id,
            "a": self._my_steam_id,
            "k": confirmation_key,
            "t": timestamp,
            "m": "android",
            "tag": tag_string,
        }

    def _select_trade_offer_confirmation(
        self, confirmations: list[Confirmation], trade_offer_id: str
    ) -> Confirmation:
        for confirmation in confirmations:
            confirmation_details_page = self._fetch_confirmation_details_page(
                confirmation
            )
            confirmation_id = self._get_confirmation_trade_offer_id(
                confirmation_details_page
            )
            if confirmation_id == trade_offer_id:
                return confirmation

        raise ConfirmationExpected

    @staticmethod
    def _get_confirmation_trade_offer_id(confirmation_details_page: str) -> str:
        soup = BeautifulSoup(confirmation_details_page, "html.parser")
        full_offer_id = soup.select(".tradeoffer")[0]["id"]

        return full_offer_id.split("_")[1]



from base64 import b64encode
from http import HTTPStatus

from requests import Response, Session
from rsa import PublicKey, encrypt

from .exceptions import (
    ApiException,
    CaptchaRequired,
    InvalidCredentials,
    SteampieException,
)
from .guard import generate_one_time_code
from .models import SteamUrl
from .utils import create_cookie


class LoginExecutor:
    def __init__(
        self, username: str, password: str, shared_secret: str, session: Session
    ) -> None:
        self.username = username
        self.password = password
        self.one_time_code = ""
        self.shared_secret = shared_secret
        self.session = session
        self.refresh_token = ""

    def _api_call(
        self,
        method: str,
        service: str,
        endpoint: str,
        version: str = "v1",
        params: dict | None = None,
    ) -> Response:
        url = f"{SteamUrl.API_URL}/{service}/{endpoint}/{version}"
        # All requests from the login page use the same 'Referer' and 'Origin' values
        headers = {
            "Referer": f"{SteamUrl.COMMUNITY_URL}/",
            "Origin": SteamUrl.COMMUNITY_URL,
        }

        if method.upper() == "GET":
            return self.session.get(url, params=params, headers=headers)

        if method.upper() == "POST":
            return self.session.post(url, data=params, headers=headers)

        raise ValueError("Method must be either GET or POST")

    def login(self) -> Session:
        login_response = self._send_login_request()

        if not login_response.json()["response"]:
            raise ApiException(
                "No response received from Steam API. Please try again later."
            )

        self._check_for_captcha(login_response)
        self._update_steam_guard(login_response)

        finalized_response = self._finalize_login()
        self._perform_redirects(finalized_response.json())
        self.set_sessionid_cookies()

        return self.session

    def _send_login_request(self) -> Response:
        rsa_params = self._fetch_rsa_params()
        encrypted_password = self._encrypt_password(rsa_params)
        rsa_timestamp = rsa_params["rsa_timestamp"]
        request_data = self._prepare_login_request_data(
            encrypted_password, rsa_timestamp
        )

        return self._api_call(
            "POST",
            "IAuthenticationService",
            "BeginAuthSessionViaCredentials",
            params=request_data,
        )

    def set_sessionid_cookies(self) -> None:
        community_domain = SteamUrl.COMMUNITY_URL[8:]
        store_domain = SteamUrl.STORE_URL[8:]
        community_cookie_dic = self.session.cookies.get_dict(domain=community_domain)
        store_cookie_dic = self.session.cookies.get_dict(domain=store_domain)

        for name in [
            "steamLoginSecure",
            "sessionid",
            "steamRefresh_steam",
            "steamCountry",
        ]:
            cookie = self.session.cookies.get_dict()[name]

            if name == "steamLoginSecure":
                store_cookie = create_cookie(name, store_cookie_dic[name], store_domain)
            else:
                store_cookie = create_cookie(name, cookie, store_domain)

            if name in ["sessionid", "steamLoginSecure"]:
                community_cookie = create_cookie(
                    name, community_cookie_dic[name], community_domain
                )
            else:
                community_cookie = create_cookie(name, cookie, community_domain)

            self.session.cookies.set(**community_cookie)
            self.session.cookies.set(**store_cookie)

    def _fetch_rsa_params(self, current_number_of_repetitions: int = 0) -> dict:
        self.session.post(SteamUrl.COMMUNITY_URL)
        request_data = {"account_name": self.username}
        response = self._api_call(
            "GET",
            "IAuthenticationService",
            "GetPasswordRSAPublicKey",
            params=request_data,
        )

        if response.status_code == HTTPStatus.OK and "response" in response.json():
            key_data = response.json()["response"]

            # Steam may return an empty 'response' value even if the status is 200
            if (
                "publickey_mod" in key_data
                and "publickey_exp" in key_data
                and "timestamp" in key_data
            ):
                rsa_mod = int(key_data["publickey_mod"], 16)
                rsa_exp = int(key_data["publickey_exp"], 16)
                return {
                    "rsa_key": PublicKey(rsa_mod, rsa_exp),
                    "rsa_timestamp": key_data["timestamp"],
                }

        maximal_number_of_repetitions = 5

        if current_number_of_repetitions < maximal_number_of_repetitions:
            return self._fetch_rsa_params(current_number_of_repetitions + 1)

        raise ApiException(
            f"Could not obtain rsa-key. Status code: {response.status_code}"
        )

    def _encrypt_password(self, rsa_params: dict) -> bytes:
        return b64encode(encrypt(self.password.encode("utf-8"), rsa_params["rsa_key"]))

    def _prepare_login_request_data(
        self, encrypted_password: bytes, rsa_timestamp: str
    ) -> dict:
        return {
            "persistence": "1",
            "encrypted_password": encrypted_password,
            "account_name": self.username,
            "encryption_timestamp": rsa_timestamp,
        }

    @staticmethod
    def _check_for_captcha(login_response: Response) -> None:
        if login_response.json().get("captcha_needed", False):
            raise CaptchaRequired("Captcha required")

    def _enter_steam_guard_if_necessary(self, login_response: Response) -> Response:
        if login_response.json()["requires_twofactor"]:
            self.one_time_code = generate_one_time_code(self.shared_secret)
            return self._send_login_request()

        return login_response

    @staticmethod
    def _assert_valid_credentials(login_response: Response) -> None:
        response = login_response.json()

        if response["success"]:
            return

        raise InvalidCredentials(response["message"])

    def _perform_redirects(self, response_dict: dict) -> None:
        parameters = response_dict.get("transfer_info")

        if parameters is None:
            raise SteampieException(
                "Cannot perform redirects after login, no parameters fetched"
            )

        for pass_data in parameters:
            pass_data["params"]["steamID"] = response_dict["steamID"]
            self.session.post(pass_data["url"], pass_data["params"])

    def _update_steam_guard(self, login_response: Response) -> None:
        response = login_response.json()["response"]
        client_id = response["client_id"]
        steamid = response["steamid"]
        request_id = response["request_id"]
        code_type = 3
        code = generate_one_time_code(self.shared_secret)

        update_data = {
            "client_id": client_id,
            "steamid": steamid,
            "code_type": code_type,
            "code": code,
        }
        response = self._api_call(
            "POST",
            "IAuthenticationService",
            "UpdateAuthSessionWithSteamGuardCode",
            params=update_data,
        )

        if response.status_code != HTTPStatus.OK:
            raise SteampieException("Cannot update Steam guard")

        self._pool_sessions_steam(client_id, request_id)

    def _pool_sessions_steam(self, client_id: str, request_id: str) -> None:
        pool_data = {"client_id": client_id, "request_id": request_id}
        response = self._api_call(
            "POST", "IAuthenticationService", "PollAuthSessionStatus", params=pool_data
        )
        self.refresh_token = response.json()["response"]["refresh_token"]

    def _finalize_login(self) -> Response:
        sessionid = self.session.cookies["sessionid"]
        redir = f"{SteamUrl.COMMUNITY_URL}/login/home/?goto="
        finalized_data = {
            "nonce": self.refresh_token,
            "sessionid": sessionid,
            "redir": redir,
        }

        return self.session.post(
            SteamUrl.LOGIN_URL + "/jwt/finalizelogin", data=finalized_data
        )
